/*
 * Copyright (C) 2014-2016 CERN (www.cern.ch)
 * Author: Federico Vaga <federico.vaga@cern.ch>
 *         inspired by a draft of Tomasz Wlostowski <tomasz.wlostowski@cern.ch>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <stdint.h>
#include <libmockturtle.h>
#include <libwrtd-internal.h>

static const char * const wrtd_errors[] = {
	"Received an invalid answer from white-rabbit-node-code CPU",
	"Cannot read channel/trigger state",
	"You are using an invalid binary",
	"Invalid dead time value",
	"Invalid delay value",
	"Invalid trigger identifier",
	"Invalid channel number",
	"Function not yet implemented",
	"Received an invalid trigger entry",
	"Received an invalid hash entry",
	"Received an invalid hash chain",
	"Received an invalid trigger handle",
	"Trigger not found",
	"No trigger condition",
	"Invalid pulse width",
	"Invalid input real-time application version",
	"Invalid output real-time application version"
};


/**
 * It returns a string messages corresponding to a given error code. If
 * it is not a libwrtd error code, it will run trtl_strerror()
 * @param[in] err error code
 * @return a message error
 */
const char *wrtd_strerror(int err)
{
	if (err < EWRTD_INVALID_ANSWER_ACK || err >= __EWRTD_MAX_ERROR_NUMBER)
		return trtl_strerror(err);

	return wrtd_errors[err - EWRTD_INVALID_ANSWER_ACK];
}


/**
 * It initializes the WRTD library. It must be called before doing
 * anything else. If you are going to load/unload WRTD devices, then
 * you have to un-load (wrtd_exit()) e reload (wrtd_init()) the library.
 *
 * This library is based on the libmockturtle, so internally, this function also
 * run trtl_init() in order to initialize the WRNC library.
 * @return 0 on success, otherwise -1 and errno is appropriately set
 */
int wrtd_init(void)
{
	int err;

	err = trtl_init();
	if (err)
		return err;

	return 0;
}


/**
 * It releases the resources allocated by wrtd_init(). It must be called when
 * you stop to use this library. Then, you cannot use functions from this
 * library.
 */
void wrtd_exit(void)
{
	trtl_exit();
}


static int __wrtd_load(struct wrtd_desc *wrtd, char *fw,
		       unsigned int cpu, uint16_t id)
{
	struct trtl_fw_version version;
	int err;

	err = trtl_cpu_load_application_file(wrtd->trtl, cpu, fw);
	if (err)
		return err;

	err = trtl_cpu_enable(wrtd->trtl, cpu);
	if (err)
		return err;

        /* Do a ping, but wait for 5sec for the FW to be ready.  */
	err = trtl_fw_ping_timeout(wrtd->trtl, cpu, 0, 5000);
	if (err)
		return err;

	err = trtl_fw_version(wrtd->trtl, cpu, 0, &version);
	if (err)
		return err;

	if (version.rt_id != id) {
		errno = EWRTD_INVALID_IN_APP;
		return -1;
	}

	return 0;
}


/**
 * It loads the input firmare (TDC)
 * @param[in] dev device token
 * @param[in] fw path to firmware
 * @return 0 on success, otherwise -1 and errno is appropriately set
 */
int wrtd_in_load(struct wrtd_node *dev, char *fw)
{
	struct wrtd_desc *wrtd = (struct wrtd_desc *)dev;

	return __wrtd_load(wrtd, fw, WRTD_CPU_TDC, WRTD_IN_RT_ID);
}


/**
 * It loads the output firmare (FD)
 * @param[in] dev device token
 * @param[in] fw path to firmware
 * @return 0 on success, otherwise -1 and errno is appropriately set
 */
int wrtd_out_load(struct wrtd_node *dev, char *fw)
{
	struct wrtd_desc *wrtd = (struct wrtd_desc *)dev;

	return __wrtd_load(wrtd, fw, WRTD_CPU_FD, WRTD_OUT_RT_ID);
}


/**
 * It opens and initialize the configuration for the given device
 * @param[in] device_id device identifier
 * @return It returns an anonymous wrtd_node structure on success.
 *         On error, NULL is returned, and errno is set appropriately.
 */
struct wrtd_node *wrtd_open(uint32_t device_id)
{
	struct wrtd_desc *wrtd;

	wrtd = malloc(sizeof(struct wrtd_desc));
	if (!wrtd)
		return NULL;

	wrtd->trtl = trtl_open_by_id(device_id);
	if (!wrtd->trtl)
		goto out;

	wrtd->dev_id = device_id;

	return (struct wrtd_node *)wrtd;

out:
	free(wrtd);
	return NULL;
}


/**
 * It closes a WRTD device opened with one of the following function:
 * wrtd_open_by_lun(), wrtd_open_by_fmc()
 * @param[in] dev device token
 */
void wrtd_close(struct wrtd_node *dev)
{
	struct wrtd_desc *wrtd = (struct wrtd_desc *)dev;

	trtl_close(wrtd->trtl);
	free(wrtd);
	dev = NULL;
}


/**
 * It returns the WRNC token in order to allows users to run
 * functions from the WRNC library
 * @param[in] dev device token
 * @return the WRNC token
 */
struct trtl_dev *wrtd_get_trtl_dev(struct wrtd_node *dev)
{
	struct wrtd_desc *wrtd = (struct wrtd_desc *)dev;

	return (struct trtl_dev *)wrtd->trtl;
}


/**
 * It converts the white rabbit time stamp to a pico seconds
 * @param[in] ts time-stamp
 * @param[out] pico pico-seconds
 */
void wrtd_ts_to_pico(struct wr_timestamp *ts, uint64_t *pico)
{
	uint64_t p;

	p = ts->frac * 8000 / 4096;
	p += (uint64_t) ts->ticks * 8000LL;
	p += ts->seconds * (1000ULL * 1000ULL * 1000ULL * 1000ULL);
	*pico = p;
}


/**
 * It converts a pico seconds integer into a white rabbit time stamp
 * @param[in] pico pico-seconds
 * @param[out] ts time-stamp
 */
void wrtd_pico_to_ts(uint64_t *pico, struct wr_timestamp *ts)
{
	uint64_t p = *pico;

	ts->seconds = p / (1000ULL * 1000ULL * 1000ULL * 1000ULL);
	p %= (1000ULL * 1000ULL * 1000ULL * 1000ULL);
	ts->ticks = p / 8000;
	p %= 8000;
	ts->frac = p * 4096 / 8000;
}


/**
 * It converts a white rabbit time stamp to seconds and pico-seconds
 * @param[in] ts time-stamp
 * @param[out] sec seconds
 * @param[out] pico pico-seconds
 */
void wrtd_ts_to_sec_pico(struct wr_timestamp *ts, uint64_t *sec, uint64_t *pico)
{
	*sec = ts->seconds;
	*pico = ts->frac * 8000 / 4096;
	*pico += (uint64_t) ts->ticks * 8000LL;
}


/**
 * It converts a white rabbit time stamp to seconds and pico-seconds
 * @param[in] sec seconds
 * @param[in] pico pico-seconds
 * @param[out] ts time-stamp
 */
void wrtd_sec_pico_to_ts(uint64_t sec, uint64_t pico, struct wr_timestamp *ts)
{
	ts->seconds = sec;
	ts->ticks = pico / 8000;
	ts->frac = (pico % 8000) * 4096 / 8000;
}

/**
 * Substract two timestamp. Returns -1 if the result would be negative.
 * @param[in] l left operand
 * @param[in] r right operand
 * @param[out] res result
 */
int wrtd_ts_sub(struct wr_timestamp *res,
                struct wr_timestamp *l, struct wr_timestamp *r)
{
        int carry;
        uint32_t rticks;
        uint64_t rsecs;

        carry = (l->frac < r->frac);
        res->frac = (carry ? 4096 : 0) + l->frac - r->frac;

        rticks = r->ticks + (carry ? 1 : 0);
        carry = l->ticks < rticks;
        res->ticks = (carry ? 125000000 : 0) + l->ticks - rticks;

        rsecs = r->seconds + (carry ? 1 : 0);
        if (l->seconds < rsecs)
                return - 1;
        res->seconds = l->seconds - rsecs;

        return 0;
}
