/*
 * Copyright (C) 2014-2016 CERN (www.cern.ch)
 * Author: Federico Vaga <federico.vaga@cern.ch>
 *         inspired by a draft of Tomasz Wlostowski <tomasz.wlostowski@cern.ch>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <libmockturtle.h>
#include "libwrtd-internal.h"


/**
 * It validates the answer of a synchronous message
 * @param[in] msg message to validate
 * @return 0 if it is valid, -1 otherwise and errno is appropriately set
 */
int wrtd_validate_acknowledge(struct trtl_msg *msg)
{
	if (msg->hdr.len != 2 || msg->data[0] != WRTD_REP_ACK_ID) {
		errno = EWRTD_INVALID_ANSWER_ACK;
		return -1;
	}

	return 0;
}


/**
 * It extracts a wr_timestamp from a given buffer (arriving from a real-time
 * application)
 * @param[in] buf answer of the real time application
 * @param[in] offset offset of the timestamp inside the answer buffer
 * @param[out] ts where write the wr_timestamp
 */
void unbag_ts(uint32_t *buf, int offset, struct wr_timestamp *ts)
{
    ts->seconds = buf[offset];
    ts->ticks = buf[offset + 1];
    ts->frac = buf[offset + 2];
}


/**
 * It compares two triggers id. The output is the same of memcmp(2)
 * @param[in] id1 first id to compare
 * @param[in] id2 second id to compare
 * @return like memcmp(2)
 */
int wrtd_trig_id_cmp(struct wrtd_trig_id *id1, struct wrtd_trig_id *id2)
{
	return memcmp(id1, id2, sizeof(struct wrtd_trig_id));
}


/*
 * Internal helper to send and receive synchronous messages to/from the WRNC
 */
int wrtd_send_and_receive_sync(struct wrtd_desc *wrtd,
			       struct trtl_msg *msg,
			       int cpu)
{
	int err;

	err = trtl_msg_sync(wrtd->trtl, cpu, 0,
			    msg, msg, WRTD_DEFAULT_TIMEOUT);

	return err < 0 ? err : 0; /* ignore timeout */
}

/**
 * It performs a simple request to a given core which will only answer
 * with an ACK
 */
int wrtd_trivial_request(struct wrtd_node *dev,
			 struct trtl_msg *request_msg,
			 int cpu)
{
	struct wrtd_desc *wrtd = (struct wrtd_desc *)dev;
	int err;

	err = wrtd_send_and_receive_sync(wrtd, request_msg, cpu);
        if (err)
		return err;

	return wrtd_validate_acknowledge(request_msg);
}
