/*
 * Copyright (C) 2015-2016 CERN (www.cern.ch)
 * Author: Federico Vaga <federico.vaga@cern.ch>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <poll.h>
#include <libmockturtle.h>
#include <libwrtd-internal.h>

/**
 * It returns a human readable string that describe a given log level
 * @param[in] lvl log level
 * @return a string if the log level is mapped, otherwise an empty string
 */
const char *wrtd_strlogging(enum wrtd_log_level lvl)
{
	switch (lvl & WRTD_LOG_LEVEL_MASK) {
	case WRTD_LOG_NOTHING:
		return "off";
	case WRTD_LOG_RAW:
		return "raw";
	case WRTD_LOG_SENT:
		return "sent";
	case WRTD_LOG_PROMISC:
		return "promiscuous";
	case WRTD_LOG_FILTERED:
		return "filtered";
	case WRTD_LOG_EXECUTED:
		return "executed";
	case WRTD_LOG_MISSED:
		return "missed";
	case WRTD_LOG_ALL:
		return "all";
	}

	return "n/a";
}


/**
 * It returns the full string describing the log_level in use
 * @param[out] buf where write the string
 * @param[in] log_level the log level to describe
 */
void wrtd_strlogging_full(char *buf, uint32_t log_level)
{
	enum wrtd_log_level lvl;

	if (!log_level) { /* No log level */
		strcpy(buf, wrtd_strlogging(log_level));
		return;
	}

	strcpy(buf,"");
	for (lvl = 0x1; lvl <= WRTD_LOG_MISSED; lvl <<= 1) {
		if (lvl & log_level) {
			strcat(buf, wrtd_strlogging(lvl));
			strcat(buf, " ");
		}
	}
}


/**
 * It converts a given logging string into a log_level
 * @param[in] log string log level
 * @return the correspondent log level enum
 */
enum wrtd_log_level wrtd_strlogging_to_level(char *log)
{
	if(!strcmp(log, "all"))
		return WRTD_LOG_ALL;
	if(!strcmp(log, "promiscuous"))
		return WRTD_LOG_PROMISC;
	if(!strcmp(log, "raw"))
		return WRTD_LOG_RAW;
	if(!strcmp(log, "executed"))
		return WRTD_LOG_EXECUTED;
	if(!strcmp(log, "missed"))
		return WRTD_LOG_MISSED;
	if(!strcmp(log, "sent"))
		return WRTD_LOG_SENT;
	if(!strcmp(log, "filtered"))
		return WRTD_LOG_FILTERED;

	return WRTD_LOG_NOTHING;
}


/**
 * It reads one or more log entry from a given hmq_log. The user of this
 * function must check that the hmq_log used correspond to a logging interface
 * @param[in] hmq_log logging HMQ.
 * @param[out] log log message
 * @param[in] count number of messages to read
 * @param[in] poll_timeout poll(2) timeout argument. Negative means infinite.
 * @return number of read messages on success (check errno if it returns less
 *         messages than expected), -1 on error and errno is set appropriately
 */
int wrtd_log_read(struct wrtd_node *dev,
		  struct wrtd_log_entry *log,
		  int count, int poll_timeout)
{
	struct wrtd_desc *wrtd = (struct wrtd_desc *)dev;
	struct wrtd_log_entry *cur = log;
	struct trtl_msg msg;
#define __WRTD_POLL_N 2
	struct polltrtl p[__WRTD_POLL_N];
	int remaining = count;
	int n_read = 0, ret, i;

	for (i = 0; i < __WRTD_POLL_N; ++i) {
		p[i].trtl = wrtd->trtl;
		p[i].idx_hmq = 0;
		p[i].events = POLLIN;
	}
	p[0].idx_cpu = WRTD_CPU_TDC;
	p[1].idx_cpu = WRTD_CPU_FD;

	/* Clean up errno to be able to distinguish between error cases and
	   normal behaviour when the function return less messages
	   than expected */
	errno = 0;
	while (remaining > 0) {
		ret = trtl_msg_poll(p, 2, poll_timeout);
		if (ret <= 0)
			break;

		for (i = 0; i < __WRTD_POLL_N; ++i) {
			if (!(p[i].revents & POLLIN))
				continue;
			ret = trtl_msg_async_recv(p[i].trtl,
						  p[i].idx_cpu,
						  p[i].idx_hmq,
						  &msg, 1);
			if (ret <= 0)
				break;
			memcpy(cur, msg.data, sizeof(struct wrtd_log_entry));
			cur->type |= (p[i].idx_cpu << WRTD_LOG_CPU_SHIFT);
			remaining--;
			n_read++;
			cur++;
		}
	}

#undef __WRTD_POLL_N
	return (n_read > 0 || errno == 0 ? n_read : -1);
}


/**
 * @param[in] dev device token
 * @param[in] channel 0-based channel index
 * @param[in] log_level log level to apply to the logging messages
 * @return 0 on success, -1 on error and errno is set appropriately
 */
static int wrtd_log_level_set(struct wrtd_node *dev, unsigned int channel,
			      uint32_t log_level, int core)
{
	struct wrtd_desc *wrtd = (struct wrtd_desc *)dev;
	struct trtl_tlv tlv;
	struct wrtd_out_channel ochan;
	struct wrtd_in_channel ichan;
	int err;

	switch (core) {
	case WRTD_CPU_FD:
		if (channel >= FD_NUM_CHANNELS) {
			errno = EWRTD_INVALID_CHANNEL;
			return -1;
		}

		tlv.type = OUT_STRUCT_CHAN_0 + channel;
		tlv.size = WRTD_OUT_CHANNEL_PUBLIC_SIZE;
		tlv.buf = &ochan;
		break;
	case WRTD_CPU_TDC:
		if (channel >= TDC_NUM_CHANNELS) {
			errno = EWRTD_INVALID_CHANNEL;
			return -1;
		}
		tlv.type = IN_STRUCT_CHAN_0 + channel;
		tlv.size = sizeof(struct wrtd_in_channel);
		tlv.buf = &ichan;
		break;
	default:
		abort();
	}

	err = trtl_fw_buffer_get(wrtd->trtl, core, 0, &tlv, 1);
	if (err)
		return err;

	switch(core) {
	case WRTD_CPU_FD:
		ochan.config.log_level = log_level;
		break;
	case WRTD_CPU_TDC:
		ichan.config.log_level = log_level;
		break;
	default:
		abort();
	}

	return trtl_fw_buffer_set(wrtd->trtl, core, 0, &tlv, 1);
}


/**
 * It sets the logging level for an output channel
 * @param[in] dev device token
 * @param[in] output index (0-based) of output channel
 * @param[in] log_level log level to apply to the logging messages
 * @return 0 on success, -1 on error and errno is set appropriately
 */
int wrtd_out_log_level_set(struct wrtd_node *dev, unsigned int output,
			   uint32_t log_level)
{
	return wrtd_log_level_set(dev, output, log_level, WRTD_CPU_FD);
}


/**
 * It gets the logging level for an output channel
 * @param[in] dev device token
 * @param[in] output index (0-based) of output channel
 * @param[out] log_level current log level used by the Real-Time application
 * @return 0 on success, -1 on error and errno is set appropriately
 */
int wrtd_out_log_level_get(struct wrtd_node *dev, unsigned int input,
			   uint32_t *log_level)
{
	struct wrtd_output_state state;
	int err;

	err = wrtd_out_state_get(dev, input, &state);
	if (err)
		return err;

	*log_level = state.log_level;

	return 0;
}


/**
 * It sets the logging level for an input channel
 * @param[in] dev device token
 * @param[in] input index (0-based) of input channel
 * @param[in] log_level log level to apply to the logging messages
 * @return 0 on success, -1 on error and errno is set appropriately
 */
int wrtd_in_log_level_set(struct wrtd_node *dev, unsigned int input,
			  uint32_t log_level)
{
	return wrtd_log_level_set(dev, input, log_level, WRTD_CPU_TDC);
}


/**
 * It gets the logging level for an input channel
 * @param[in] dev device token
 * @param[in] input index (0-based) of input channel
 * @param[out] log_level current log level used by the Real-Time application
 * @return 0 on success, -1 on error and errno is set appropriately
 */
int wrtd_in_log_level_get(struct wrtd_node *dev, unsigned int input,
			  uint32_t *log_level)
{
	struct wrtd_input_state state;
	int err;

	err = wrtd_in_state_get(dev, input, &state);
	if (err)
		return err;

	*log_level = state.log_level;
	return 0;
}
