#include <stdarg.h>
#include <stddef.h>
#include <stdlib.h>
#include <assert.h>
#include <errno.h>
#include "libmockturtle.h"
#include "libwrtd.h"
#include "libwrtd-private.h"

#define TRACE

struct rule_map {
        struct wrtd_lib_rule *rule;
        int cpu;
        int idx;
};

struct device_map {
        const struct wrtd_mt_device *dev;
        /* CPU that drives this device.  */
        unsigned int cpu;
        /* Device index on the cpu.  */
        unsigned int dev_idx;
        /* Enabled/used channels.  */
        uint32_t ch_enable;
};

struct conf_map {
        unsigned nbr_rules;
        struct rule_map *rules;
        unsigned nbr_devs;
        struct device_map *devs;
        unsigned free_rule_slots[WRTD_MAX_CPUS];
};


static enum wrtd_status wrtd_reconfigure_alloc_map(struct wrtd_dev *wrtd,
                                                   struct conf_map *map)
{
        enum wrtd_status status;
        unsigned cpu;
        unsigned n;
        unsigned i;

        /* 1. Count number of rules.  */
        status = wrtd_attr_get_rule_count(wrtd, &map->nbr_rules);
        WRTD_RETURN_IF_ERROR(status);

        /* Count number of devices.  */
        n = 0;
        for (cpu = 0; cpu < wrtd->config->nbr_cpus; cpu++)
                n += wrtd->config->cpus[cpu].nbr_devices;
        map->nbr_devs = n;

        /* Allocate maps.  */
        map->rules = malloc(map->nbr_rules * sizeof(struct rule_map));
        if (map->rules == NULL)
                return wrtd_return_error(wrtd, WRTD_ERROR_OUT_OF_MEMORY,
                                         "cannot allocate map for %u rules",
                                         map->nbr_rules);

        map->devs = malloc(map->nbr_devs * sizeof(struct device_map));
        if (map->devs == NULL) {
                free(map->rules);
                map->rules = NULL;
                return wrtd_return_error(wrtd, WRTD_ERROR_OUT_OF_MEMORY,
                                         "cannot allocate map for %u devs",
                                         map->nbr_devs);
        }

        /* Fill rules.  */
        n = 0;
        for (i = 0; i < wrtd->nbr_rules; i++) {
                if (wrtd_id_null(wrtd->rules[i].rule.conf.id))
                        continue;
                map->rules[n].rule = &wrtd->rules[i];
                map->rules[n].cpu = -1;
                map->rules[n].idx = -1;
                n++;
        }
        assert (n == map->nbr_rules);

        /* No rules placed.  */
        for (i = 0; i < wrtd->nbr_cpus; i++)
                map->free_rule_slots[i] = wrtd->roots[i].nbr_rules;

        /* Fill devs.  */
        n = 0;
        for (cpu = 0; cpu < wrtd->config->nbr_cpus; cpu++)
                for (i = 0; i < wrtd->config->cpus[cpu].nbr_devices; i++) {
                        map->devs[n].dev = wrtd->config->cpus[cpu].devices[i];
                        map->devs[n].cpu = cpu;
                        map->devs[n].dev_idx = i;
                        map->devs[n].ch_enable = 0;
                        n++;
                }

        return WRTD_SUCCESS;
}

static enum wrtd_status wrtd_find_channel(struct wrtd_dev *wrtd,
                                          const struct conf_map *map,
                                          const char *id, enum wrtd_dir dir,
                                          unsigned *res_dev,
                                          unsigned *res_ch)
{
        unsigned i;
        unsigned j;

        for (i = 0; i < map->nbr_devs; i++) {
                struct device_map *d = &map->devs[i];
                for (j = 0; j < d->dev->nbr_chs; j++)
                        if (d->dev->dir == dir
                            && wrtd_id_eq(d->dev->channels[j], id)) {
                                d->ch_enable |= 1 << j;
                                *res_dev = i;
                                *res_ch = (d->dev_idx << 4) + j;
                                return WRTD_SUCCESS;
                        }
        }

        return WRTD_ERROR_RESOURCE_UNKNOWN;
}

/* Assign RULE_NUM to CPU.  */
static enum wrtd_status wrtd_reconfigure_place_rule(struct wrtd_dev *wrtd,
                                                    struct conf_map *map,
                                                    unsigned rule_num,
                                                    unsigned cpu)
{
        assert(map->rules[rule_num].cpu == -1);

        if (map->free_rule_slots[cpu] == 0)
                return wrtd_return_error(wrtd, WRTD_ERROR_RULE_OUT_OF_RESOURCES,
                                         "no more rule slots on cpu %u", cpu);

        map->rules[rule_num].cpu = cpu;
        map->free_rule_slots[cpu]--;

        if (map->rules[rule_num].rule->rule.conf.dest_cpu == cpu)
                map->rules[rule_num].rule->rule.conf.dest_cpu = WRTD_DEST_CPU_LOCAL;

        return WRTD_SUCCESS;
}

static enum wrtd_status wrtd_reconfigure_place(struct wrtd_dev *wrtd,
                                               struct conf_map *map)
{
        enum wrtd_status status;
        unsigned i;
        unsigned cpu;
        unsigned disable_idx;
        unsigned cpu_offset;

        /* Assign enabled rules to cpu.  */
        for (i = 0; i < map->nbr_rules; i++) {
                struct rule_map *m = &map->rules[i];
                struct wrtd_rule *rule = &m->rule->rule;
                unsigned dev;
                unsigned ch;
                unsigned cpu_affinity;

                /* Skip disabled rules.  */
                if (!rule->conf.enabled)
                        continue;

                /* Find reserved input channel.  */
                status = wrtd_find_channel(wrtd, map, rule->conf.source_id,
                                           wrtd_input, &dev, &ch);
                if (status == WRTD_SUCCESS) {
                        cpu_affinity = map->devs[dev].cpu;
                }
                else {
                        status = wrtd_find_alarm(
                                wrtd, rule->conf.source_id, &ch);
                        if (status == WRTD_SUCCESS) {
                                cpu_affinity = wrtd->alarms[ch].cpu;
                        }
                        else {
                                cpu_affinity = wrtd->config->rx_cpu;
                        }
                }

                /* Find reserved output channel.  */
                status = wrtd_find_channel(wrtd, map, rule->conf.dest_id,
                                           wrtd_output, &dev, &ch);
                if (status == WRTD_SUCCESS) {
                        rule->conf.dest_cpu = map->devs[dev].cpu;
                        rule->conf.dest_ch = ch;
                }
                else {
                        rule->conf.dest_cpu = wrtd->config->tx_cpu;
                        rule->conf.dest_ch = WRTD_DEST_CH_NET;
                }

                /* Place rules according to input affinity.  */
                status = wrtd_reconfigure_place_rule
                        (wrtd, map, i, cpu_affinity);
                WRTD_RETURN_IF_ERROR(status);
        }

        /* Assign indexes, handle collisions.  */
        disable_idx = 0;
        cpu_offset = 0;
        for(cpu = 0; cpu < wrtd->nbr_cpus; cpu++) {
                const unsigned nbr_cpu_rules = wrtd->roots[cpu].nbr_rules;
                /* SLOTS[N] contains the index of the rule in map.  */
                int *slots;
                int free_slot_idx;

                /* Allocate slots array.  */
                slots = (int *)calloc(nbr_cpu_rules, sizeof(int));
                if (slots == NULL)
                        return wrtd_return_error
                                (wrtd, WRTD_ERROR_OUT_OF_MEMORY,
                                 "cannot allocate slots for %u rules",
                                 nbr_cpu_rules);

                /* Marks slot as unused.  */
                for (i = 0; i < nbr_cpu_rules; i++)
                        slots[i] = -1;

                /* Compute hash and place heads.  */
                for (i = 0; i < map->nbr_rules; i++) {
                        struct rule_map *m = &map->rules[i];
                        unsigned int hash;
                        if (m->cpu != cpu)
                                continue;

                        hash = wrtd_id_hash(m->rule->rule.conf.source_id)
                                % nbr_cpu_rules;

                        m->idx = cpu_offset + hash;
                        m->rule->rule.conf.hash_chain = -1;
                        if (slots[hash] == -1)
                                slots[hash] = i;
                }

                /* Place collisions.  */
                free_slot_idx = 0;
                for (i = 0; i < map->nbr_rules; i++) {
                        struct rule_map *m = &map->rules[i];
                        unsigned head_idx;

                        if (m->cpu != cpu)
                                continue;
                        head_idx = m->idx - cpu_offset;
                        assert(head_idx < nbr_cpu_rules);

                        /* Skip it if already placed.  */
                        if (slots[head_idx] == i)
                                continue;

                        /* Find next free slot (there must be one).  */
                        while (slots[free_slot_idx] != -1) {
                                free_slot_idx++;
                                assert(free_slot_idx < nbr_cpu_rules);
                        }
                        m->idx = cpu_offset + free_slot_idx;
                        slots[free_slot_idx] = i;

                        /* Chain.  */
                        m->rule->rule.conf.hash_chain =
                                map->rules[head_idx].rule->rule.conf.hash_chain;
                        map->rules[head_idx].rule->rule.conf.hash_chain =
                                free_slot_idx;

                        free_slot_idx++;
                }

                /* Place disabled rules.  */
                while(1) {
                        /* Find next disable rule.  */
                        while (disable_idx < map->nbr_rules
                               && map->rules[disable_idx].rule->rule.conf.enabled)
                                disable_idx++;
                        if (disable_idx >= map->nbr_rules)
                                break;

                        /* Find next free slot.  */
                        while (slots[free_slot_idx] != -1) {
                                free_slot_idx++;
                                assert(free_slot_idx < nbr_cpu_rules);
                        }

                        map->rules[disable_idx].idx =
                                cpu_offset + free_slot_idx;
                        free_slot_idx++;
                        disable_idx++;
                }

                free(slots);
                cpu_offset += nbr_cpu_rules;
        }

        return WRTD_SUCCESS;
}

static enum wrtd_status wrtd_reconfigure_write(struct wrtd_dev *wrtd,
                                               struct conf_map *map,
                                               struct wrtd_lib_rule *new_rules)
{
        enum wrtd_status status;
        struct wrtd_lib_rule *old_rules;
        unsigned i;

        old_rules = wrtd->rules;

        /* Copy rules.  */
        for(i = 0; i < map->nbr_rules; i++)
                new_rules[map->rules[i].idx].rule = map->rules[i].rule->rule;

        /* Freeze all cpus.  */
        /* TODO.  */

        /* For each rule: load stats if different.  */
        /* TODO.  */

        wrtd->rules = new_rules;

        for(i = 0; i < wrtd->nbr_rules; i++) {
                if (wrtd_id_null(new_rules[i].rule.conf.id)
                    && wrtd_id_null(old_rules[i].rule.conf.id))
                        continue;
                if (wrtd_id_eq(new_rules[i].rule.conf.id,
                               old_rules[i].rule.conf.id)
                    && (new_rules[i].rule.conf.enabled
                        == old_rules[i].rule.conf.enabled))
                        continue;
                status = wrtd_write_rule(wrtd, i);
                /* FIXME: inconsistent state, free old_rules ?  */
                WRTD_RETURN_IF_ERROR(status);
        }
        free(old_rules);

        for(i = 0; i < map->nbr_devs; i++) {
                struct device_map *dm = &map->devs[i];
                status = dm->dev->configure(wrtd, dm->cpu, dm->dev_idx,
                                            dm->ch_enable);
                /* FIXME: inconsistent state.  */
                WRTD_RETURN_IF_ERROR(status);
        }

        return WRTD_SUCCESS;
}

enum wrtd_status wrtd_reconfigure(struct wrtd_dev *wrtd)
{
        enum wrtd_status status;
        struct conf_map map;
        struct wrtd_lib_rule *new_rules;

        /* Load rules and alarms.  */
        status = wrtd_fill_rules(wrtd);
        WRTD_RETURN_IF_ERROR(status);

        status = wrtd_fill_alarms(wrtd);
        WRTD_RETURN_IF_ERROR(status);

        /* 1. Count number of rules.  */
        status = wrtd_reconfigure_alloc_map(wrtd, &map);
        WRTD_RETURN_IF_ERROR(status);

        status = wrtd_reconfigure_place(wrtd, &map);

        if (status == WRTD_SUCCESS)
                status = wrtd_alloc_rules(wrtd, &new_rules);

        if (status == WRTD_SUCCESS)
                status = wrtd_reconfigure_write(wrtd, &map, new_rules);

        if (status != WRTD_SUCCESS)
                free(new_rules);
        free(map.rules);
        free(map.devs);

        return status;
}
