/*
 * Copyright (C) 2018 CERN (www.cern.ch)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * @TODO software trigger test
 */

#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <time.h>
#include <inttypes.h>

#include <libmockturtle.h>
#include <libwrtd.h>

#define AssertIntEquals(VAL, EXPR) do \
        {                                                               \
        int val = (VAL);                                                \
        int expr = (EXPR);                                              \
        if (val != expr)  {                                             \
                fprintf(stderr,                                         \
                        "%s:%d: unexpected value (got %d, expect %d)\n", \
                        __FILE__, __LINE__, expr, val);                 \
                exit (1);                                               \
        }                                                               \
        } while (0)

#define AssertPtrNotNull(EXPR) do \
        {                         \
        void *ptr = (EXPR); \
        if (ptr == NULL) { \
                fprintf(stderr,                                         \
                        "%s:%d: unexpected null pointer\n",             \
                        __FILE__, __LINE__);                            \
                exit (1);                                               \
        }                                                               \
        } while (0)


static struct wrtd_node *wrtd;
static int channel;

static void setup(int devid)
{
        const int with_trigger = 0;
	struct wrtd_trig_id tid_1 = {0, 0, 1};

	wrtd = wrtd_open(devid);
	AssertPtrNotNull(wrtd);

        AssertIntEquals(0, wrtd_in_global_counters_reset(wrtd));
        AssertIntEquals(0, wrtd_in_counters_reset(wrtd, channel));

	/* Assign trigger input */
        if (with_trigger) {
                AssertIntEquals(0, wrtd_in_trigger_assign(wrtd, channel, &tid_1));
                AssertIntEquals(0, wrtd_in_trigger_mode_set(wrtd, channel, WRTD_TRIGGER_MODE_AUTO));
        }

	AssertIntEquals(0, wrtd_in_dead_time_set(wrtd, channel, 80000000));

	/* Enable and Arm */
        if (with_trigger) {
                AssertIntEquals(0, wrtd_in_arm(wrtd, channel, 1));
        }
	AssertIntEquals(0, wrtd_in_enable(wrtd, channel, 1));
}

static void print_ns(uint64_t ns)
{
        printf ("%03llu,%03llu,%03lluns",
                (ns / (1000LL * 1000)),
                (ns / 1000LL) % 1000,
                ns % 1000ULL);
}

static void print_ts(struct wr_timestamp ts)
{
        printf ("%lu:", (unsigned long)ts.seconds);
        print_ns(ts.ticks * 8ULL + (ts.frac * 8 / 4096));
        printf("+%03df", ts.frac & 0x1ff);
}

static void disp_timing(int count)
{
	struct wrtd_input_state sti;
        struct wr_timestamp last_ts;
        uint32_t n_pulse;
        int i;
        uint64_t pico;

        do
                AssertIntEquals(0, wrtd_in_state_get(wrtd, channel, &sti));
        while (sti.total_pulses == 0);
        n_pulse = sti.total_pulses;
        last_ts = sti.last_pulse;

        for (i = 1; i <= count; i++) {
                /* Get stats.  */
                do
                        AssertIntEquals(0, wrtd_in_state_get(wrtd, channel, &sti));
                while (sti.total_pulses == n_pulse);

                if (sti.total_pulses == n_pulse + 1) {
                        struct wr_timestamp d;

                        AssertIntEquals(0, wrtd_ts_sub(&d,
                                                       &sti.last_pulse,
                                                       &last_ts));
                        wrtd_ts_to_pico(&d, &pico);
                        printf("Delay: %"PRIu64"ps ", pico);
                        print_ts(d);
                        printf(", abs: ");
                        print_ts(sti.last_pulse);
                        printf("\n");
                }
                else if (sti.total_pulses > n_pulse)
                        printf ("Miss!\n");
                else
                        printf ("??? go backward\n");
                n_pulse = sti.total_pulses;
                last_ts = sti.last_pulse;
        }
}

static uint64_t measure(int count)
{
	struct wrtd_input_state sti;
        struct wr_timestamp last_ts;
        uint32_t n_pulse;
        uint64_t sum;

        do
          AssertIntEquals(0, wrtd_in_state_get(wrtd, channel, &sti));
        while (sti.total_pulses == 0);
        n_pulse = sti.total_pulses;
        last_ts = sti.last_pulse;

        /* Get stats.  */
        do
                AssertIntEquals(0, wrtd_in_state_get(wrtd, channel, &sti));
        while (sti.total_pulses < n_pulse + count);

        if (sti.total_pulses == n_pulse + count) {
                struct wr_timestamp d;

                AssertIntEquals(0, wrtd_ts_sub(&d,
                                               &sti.last_pulse,
                                               &last_ts));
                wrtd_ts_to_pico(&d, &sum);
                return sum / count;
        }
        else if (sti.total_pulses > n_pulse) {
                printf ("Miss!\n");
                return 0;
        }
        else {
                printf ("??? go backward\n");
                return 0;
        }
}

static void disp_measure(int count)
{
        uint64_t period;

        period = measure(count);
        if (period == 0)
                return;
        printf ("average delay: %"PRIu64"ps (using %d samples)\n",
                period, count);
        printf ("Freq: %f Hz\n", 1.0e12 / (double)period);
}

static void disp_graph(int count)
{
	struct wrtd_input_state sti;
        struct wr_timestamp last_ts;
        struct wr_timestamp first_ts;
        struct wr_timestamp d;
        uint32_t n_pulse;
        int i;
        uint64_t pico;
        uint64_t min_period;
        uint64_t max_period;
        uint64_t period;
        uint64_t bin_width;
        uint64_t sum;
#define CBINS 10
#define NBINS (2*CBINS + 1)
        int bins[NBINS];
        int oo_bin;
        int misses;
        int count0;
        int verbose = 0;

        count0 = count / 100;
        if (count0 < 5)
          count0 = 5;
        else if (count0 > 100)
          count0 = 100;

        do
                AssertIntEquals(0, wrtd_in_state_get(wrtd, channel, &sti));
        while (sti.total_pulses == 0);
        n_pulse = sti.total_pulses;
        last_ts = sti.last_pulse;
        first_ts = last_ts;

        max_period = 0;
        min_period = -1;
        sum = 0;

        for (i = 0; i < count0; i++) {
                /* Get stats.  */
                do
                        AssertIntEquals(0, wrtd_in_state_get(wrtd, channel, &sti));
                while (sti.total_pulses == n_pulse);

                if (sti.total_pulses == n_pulse + 1) {
                        AssertIntEquals(0, wrtd_ts_sub(&d,
                                                       &sti.last_pulse,
                                                       &last_ts));
                        wrtd_ts_to_pico(&d, &pico);
                        if (verbose) {
                                struct wr_timestamp d1;
                                uint64_t pico1;

                                //  Note: drifting due to rounding.
                                sum += pico;

                                printf("Delay: %12"PRIu64"ps ", pico);
                                print_ts(d);
                                printf(", abs: ");
                                print_ts(sti.last_pulse);
                                printf("\n");
                                AssertIntEquals(0, wrtd_ts_sub(&d1,
                                                               &sti.last_pulse,
                                                               &first_ts));
                                wrtd_ts_to_pico(&d1, &pico1);
                                if (abs(pico1 - sum) > (i + 2)) {
                                        printf("pico1: %12"PRIu64"ps, sum: %12"PRIu64"ps\n", pico1, sum);
                                        break;
                                }
                        }
                        if (pico > max_period)
                                max_period = pico;
                        if (pico < min_period)
                                min_period = pico;
                }
                else if (sti.total_pulses > n_pulse) {
                        printf ("Miss!\n");
                        return;
                }
                else {
                        printf ("??? go backward\n");
                        return;
                }
                n_pulse = sti.total_pulses;
                last_ts = sti.last_pulse;
        }

        //  Compute period.
        AssertIntEquals(0, wrtd_ts_sub(&d, &last_ts, &first_ts));
        wrtd_ts_to_pico(&d, &period);
        if (verbose) {
                printf ("last ts: ");
                print_ts (last_ts);
                printf (",  first ts: ");
                print_ts (first_ts);
                printf("\n");
                printf ("diff: ");
                print_ts(d);
                printf(" = %12"PRIu64"ps\n", period);
        }
        period /= count0;

        printf ("Mean period: %12"PRIu64"ps\n", period);
        printf ("Min period:  %12"PRIu64"ps\n", min_period);
        printf ("Max period:  %12"PRIu64"ps\n", max_period);
        printf ("Delta:       %12"PRIu64"ps\n", max_period - min_period);

        bin_width = (max_period - min_period) / CBINS;
        printf ("Bin width:   %12"PRIu64"ps = ", bin_width);
        print_ns(bin_width / 1000);
        printf ("\n");
        fflush(stdout);  /* For tee.  */

        for (i = 0; i < NBINS; i++)
                bins[i] = 0;
        oo_bin = 0;
        misses = 0;

        for (i = 1; i <= count; i++) {
                /* Get stats.  */
                do
                        AssertIntEquals(0, wrtd_in_state_get(wrtd, channel, &sti));
                while (sti.total_pulses == n_pulse);

                if (sti.total_pulses == n_pulse + 1) {
                        int n;
                        AssertIntEquals(0, wrtd_ts_sub(&d,
                                                       &sti.last_pulse,
                                                       &last_ts));
                        wrtd_ts_to_pico(&d, &pico);
                        if (verbose)
                                printf("Delay: %12"PRIu64"ps, diff: %12"PRId64"ps ",
                                       pico, (int64_t)pico - (int64_t)period);

                        /* Bins are centered around period.  */
                        if (pico >= period) {
                                n = (pico - period + bin_width / 2) / bin_width;
                                n += CBINS;
                        }
                        else {
                                n = (period - pico + bin_width / 2) / bin_width;
                                n = CBINS - n;
                        }
                        if (verbose)
                                printf("bin: %d\n", n - CBINS);
                        if (n >= 0 && n < NBINS)
                                bins[n]++;
                        else {
                                oo_bin++;
                                printf("OUT: %12"PRIu64"ps, diff: %12"PRId64"ps\n",
                                       pico, (int64_t)pico - (int64_t)period);
                        }

                }
                else if (sti.total_pulses > n_pulse) {
                        printf ("Miss!\n");
                        misses++;
                }
                else {
                        printf ("??? go backward\n");
                        return;
                }
                n_pulse = sti.total_pulses;
                last_ts = sti.last_pulse;
        }
        for (i = 0; i < NBINS; i++)
                printf("%+3d: %6d %6.01f%% %12"PRId64"ps\n",
                       i - CBINS, bins[i], (double)bins[i] * 100 / count, (i - CBINS) * bin_width);
        printf("out: %4d, misses: %d\n", oo_bin, misses);
}

static void teardown(void)
{
	/* Stop.  */
	AssertIntEquals(0, wrtd_in_enable(wrtd, channel, 0));

	/* Close */
	wrtd_close(wrtd);
}


static void help(void)
{
        printf("usage: wrtd-timetest -D DEVICE -c CHANNEL [-C COUNT]\n");
}

static int my_atoi(const char *str)
{
	int res;
	char *endptr;

	res = strtol(str, &endptr, 0);
	if (*endptr != 0)
		res = -1;
	return res;
}

int main(int argc, char **argv)
{
        int c;
        int count;
	int devid = -1;
        const char *app;

        count = 100;
	channel = -1;
        app = "time";

	while ((c = getopt (argc, argv, "hC:D:c:A:")) != -1) {
		switch (c) {
		case 'h':
		case '?':
			help();
			break;
                case 'D':
			devid = my_atoi(optarg);
                        break;
                case 'c':
			channel = my_atoi(optarg);
                        break;
                case 'C':
			count = my_atoi(optarg);
                        break;
                case 'A':
                        app = optarg;
                        break;
                }
        }

	if (devid < 0) {
		fprintf(stderr, "bad device id\n");
		exit(1);
	}
	if (channel < 0) {
		fprintf(stderr, "bad channel\n");
		exit(1);
	}
        if (count < 1) {
                fprintf(stderr, "bad count value (must be > 0)\n");
                exit(1);
        }

        setup(devid);
        if (strcmp(app, "time") == 0)
                disp_timing(count);
        else if (strcmp(app, "measure") == 0)
                disp_measure(count);
        else if (strcmp(app, "graph") == 0)
                disp_graph(count);
        else {
                printf ("Unknown application %s\n", app);
        }

        teardown();

        return 0;
}
