/*
 * Copyright (C) 2014-2016 CERN (www.cern.ch)
 * Author: Federico Vaga <federico.vaga@cern.ch>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <string.h>
#include <errno.h>
#include <inttypes.h>

#include "wrtd-ut.h"
#include "CuTest.h"

#include <libmockturtle.h>
#include <libwrtd.h>

#define MAX_SYSTEM 4
#define MAX_TRIGGER 2
#define MAX_ENTRY_CHAN (MAX_SYSTEM * MAX_TRIGGER)


static void test_out_ping(CuTest *tc)
{
	struct wrtd_node *wrtd;

	wrtd = wrtd_open(1);
	CuAssertIntEquals(tc, 0, wrtd_out_ping(wrtd));
	wrtd_close(wrtd);
}


static void test_out_channels_enable(CuTest *tc)
{
	struct wrtd_node *wrtd;
	unsigned int i, enable;

	wrtd = wrtd_open(1);

	for (i = 0; i < FD_NUM_CHANNELS; i++) {
		CuAssertIntEquals(tc, 0, wrtd_out_enable(wrtd, i, 1));
		CuAssertIntEquals(tc, 0, wrtd_out_is_enabled(wrtd, i, &enable));
		CuAssertTrue(tc, enable);
	}

	wrtd_close(wrtd);
}

static void test_out_channels_disable(CuTest *tc)
{
	struct wrtd_node *wrtd;
	unsigned int i, enable;

	wrtd = wrtd_open(1);

	for (i = 0; i < FD_NUM_CHANNELS; i++) {
		CuAssertIntEquals(tc, 0, wrtd_out_enable(wrtd, i, 0));
		CuAssertIntEquals(tc, 0, wrtd_out_is_enabled(wrtd, i, &enable));
		CuAssertIntEquals(tc, 0, enable);
	}

	wrtd_close(wrtd);
}

static void test_out_channels_arm(CuTest *tc)
{
	struct wrtd_node *wrtd;
	unsigned int i, armed;

	wrtd = wrtd_open(1);
	for (i = 0; i < FD_NUM_CHANNELS; i++) {
		CuAssertIntEquals(tc, 0, wrtd_out_arm(wrtd, i, 1));
		CuAssertIntEquals(tc, 0, wrtd_out_is_armed(wrtd, i, &armed));
		CuAssertTrue(tc, armed);
	}
	wrtd_close(wrtd);
}

static void test_out_channels_disarm(CuTest *tc)
{
	struct wrtd_node *wrtd;
	unsigned int i, armed;

	wrtd = wrtd_open(1);
	for (i = 0; i < FD_NUM_CHANNELS; i++) {
		CuAssertIntEquals(tc, 0, wrtd_out_arm(wrtd, i, 0));
		CuAssertIntEquals(tc, 0, wrtd_out_is_armed(wrtd, i, &armed));
		CuAssertIntEquals(tc, 0, armed);
	}
	wrtd_close(wrtd);
}

static void test_out_trigger_clear(CuTest *tc)
{
	struct wrtd_output_trigger_state trig[15];
	struct wrtd_node *wrtd;
        int i;

	wrtd = wrtd_open(1);

	for (i = 0; i < FD_NUM_CHANNELS; i++) {
                int num;
                int j;

		num = wrtd_out_trig_get_all(wrtd, i, trig, 15);

                for (j = 0; j < num; j++) {
                        wrtd_out_trig_unassign(wrtd, &trig[j].handle);
                        /* May have been removed as a condition of a former
                           trigger.  */
                }
        }

	wrtd_close(wrtd);
}

static void test_out_trigger_assign_one(CuTest *tc, struct wrtd_node *wrtd,
				    unsigned int chan,
				    struct wrtd_trig_id *id)
{
	struct wrtd_output_trigger_state trig;
	struct wrtd_trigger_handle h;
	unsigned int assigned;
	char msg[128];
	int ret;

	/* printf("Channel %d Trigger ID %d:%d:%d\n", */
	/* 	chan, id->system, id->source_port, id->trigger); */

	CuAssertIntEquals(tc, 0, wrtd_out_has_trigger(wrtd, chan, id, &assigned));
	CuAssertIntEquals(tc, 0, assigned);

	ret = wrtd_out_trig_assign(wrtd, chan, &h, id, NULL);
	sprintf(msg, "Assert failed - Channel %d Trigger ID %d:%d:%d - %s",
		chan, id->system, id->source_port, id->trigger, wrtd_strerror(errno));
	CuAssertIntEquals_Msg(tc, msg, 0, ret);

	ret = wrtd_out_trig_state_get_by_handle(wrtd, &h, &trig);
	sprintf(msg, "Assert failed - Channel %d Trigger ID %d:%d:%d - %s",
		chan, id->system, id->source_port, id->trigger, wrtd_strerror(errno));
	CuAssertIntEquals_Msg(tc, msg, 0, ret);

 	CuAssertIntEquals(tc, 0, memcmp(id, &trig.trigger,
					sizeof(struct wrtd_trig_id)));
	CuAssertIntEquals(tc, 0, wrtd_out_has_trigger(wrtd, chan, id, &assigned));
	CuAssertTrue(tc, assigned);
}

static void test_out_trigger_assign_all(CuTest *tc, struct wrtd_node *wrtd,
				    unsigned int chan)
{
	struct wrtd_trig_id id;
	int k, w;

	id.source_port = chan;
	for (k = MAX_SYSTEM; k > 0; k--) {
		for (w = 0; w < MAX_TRIGGER; w++) {
			id.system = k;
			id.trigger = w;
			test_out_trigger_assign_one(tc, wrtd, chan, &id);
		}
	}
}


static void test_out_trigger_assign(CuTest *tc)
{
	struct wrtd_output_trigger_state trig[15];
	struct wrtd_node *wrtd;
	unsigned int assigned;
        int i, ret;

	wrtd = wrtd_open(1);

	for (i = 0; i < FD_NUM_CHANNELS; i++) {
		ret = wrtd_out_trig_get_all(wrtd, i, trig, 15);
		CuAssertIntEquals(tc, 0, ret);
		test_out_trigger_assign_all(tc, wrtd, i);
                ret = wrtd_out_has_trigger(wrtd, i, NULL, &assigned);
                CuAssertIntEquals(tc, 0, ret);
		CuAssertTrue(tc, assigned);
		memset(&trig, 0, sizeof(trig));
		ret = wrtd_out_trig_get_all(wrtd, i, trig, MAX_ENTRY_CHAN);
		CuAssertIntEquals(tc, MAX_ENTRY_CHAN, ret);
	}

	wrtd_close(wrtd);
}


static void test_out_trigger_assign_cond_one(CuTest *tc, struct wrtd_node *wrtd,
                                             unsigned int chan,
                                             struct wrtd_trig_id *id)
{
	struct wrtd_output_trigger_state trig, cond;
	struct wrtd_trig_id condid = *id;
	struct wrtd_trigger_handle h;
	unsigned int assigned;
	char msg[128];
	int ret;

	condid.trigger += 100;
	CuAssertIntEquals(tc, 0, wrtd_out_has_trigger(wrtd, chan, id, &assigned));
	CuAssertIntEquals(tc, 0, assigned);

	ret = wrtd_out_trig_assign(wrtd, chan, &h, id, &condid);
	sprintf(msg, "Assert failed - Channel %d Trigger ID %d:%d:%d - %s",
		chan, id->system, id->source_port, id->trigger, wrtd_strerror(errno));
	CuAssertIntEquals_Msg(tc, msg, 0, ret);

	ret = wrtd_out_trig_state_get_by_handle(wrtd, &h, &trig);
	sprintf(msg, "Assert failed - Channel %d Trigger ID %d:%d:%d - %s",
		chan, id->system, id->source_port, id->trigger, wrtd_strerror(errno));
	CuAssertIntEquals_Msg(tc, msg, 0, ret);

	h.ptr_trig = h.ptr_cond;
	h.ptr_cond = 0;
	ret = wrtd_out_trig_state_get_by_handle(wrtd, &h, &cond);
	sprintf(msg, "Assert failed - Channel %d Trigger ID %d:%d:%d - %s",
		chan, id->system, id->source_port, id->trigger, wrtd_strerror(errno));
	CuAssertIntEquals_Msg(tc, msg, 0, ret);

 	CuAssertIntEquals(tc, 0, memcmp(id, &trig.trigger,
					sizeof(struct wrtd_trig_id)));
	CuAssertIntEquals(tc, 0, memcmp(&condid, &cond.trigger,
					sizeof(struct wrtd_trig_id)));
	CuAssertIntEquals(tc, 0, wrtd_out_has_trigger(wrtd, chan, id, &assigned));
	CuAssertTrue(tc, assigned);
}

static void test_out_trigger_assign_cond_all(CuTest *tc, struct wrtd_node *wrtd,
                                             unsigned int chan)
{
	struct wrtd_trig_id id;
	int k, w;

	id.source_port = chan;
	for (k = 0; k < 2; k++) {
		for (w = 0; w < 2; w++) {
			id.system = k;
			id.trigger = w;
			test_out_trigger_assign_cond_one(tc, wrtd, chan, &id);
		}
	}
}

static void test_out_trigger_assign_cond(CuTest *tc)
{
	struct wrtd_output_trigger_state trig[MAX_ENTRY_CHAN];
	struct wrtd_node *wrtd;
	unsigned int assigned;
        int i, ret;

	wrtd = wrtd_open(1);

	for (i = 0; i < FD_NUM_CHANNELS; i++) {
                ret = wrtd_out_trig_get_all(wrtd, i, trig, MAX_ENTRY_CHAN);
		CuAssertIntEquals(tc, 0, ret);
		test_out_trigger_assign_cond_all(tc, wrtd, i);
                ret = wrtd_out_has_trigger(wrtd, i, NULL, &assigned);
		CuAssertIntEquals(tc, 0, ret);
		CuAssertTrue(tc, assigned);
		memset(&trig, 0, sizeof(trig));
		ret = wrtd_out_trig_get_all(wrtd, i, trig, MAX_ENTRY_CHAN);
		CuAssertIntEquals(tc, MAX_ENTRY_CHAN, ret);
	}

	wrtd_close(wrtd);
}


static void test_out_trigger_unassign_one(CuTest *tc, struct wrtd_node *wrtd,
                                          unsigned int chan,
                                          struct wrtd_trig_id *id)
{
	struct wrtd_output_trigger_state trig;
	unsigned int assigned;
	char msg[128];
	int ret;

	CuAssertIntEquals(tc, 0, wrtd_out_has_trigger(wrtd, chan, NULL, &assigned));
	CuAssertTrue(tc, assigned);

	ret = wrtd_out_trig_state_get_by_id(wrtd, chan, id, &trig);
	sprintf(msg, "Assert failed - Chan %d Trigger ID %d:%d:%d - %s",
		chan, id->system, id->source_port, id->trigger, wrtd_strerror(errno));
	CuAssertIntEquals_Msg(tc, msg, 0, ret);

	ret = wrtd_out_trig_unassign(wrtd, &trig.handle);
	CuAssertIntEquals_Msg(tc, msg, 0, ret);

	ret = wrtd_out_trig_state_get_by_handle(wrtd, &trig.handle, &trig);
	CuAssertIntEquals_Msg(tc, msg, -1, ret);

	CuAssertIntEquals(tc, 0, wrtd_out_has_trigger(wrtd, chan, id, &assigned));
	CuAssertIntEquals(tc, 0, assigned);
}

static void test_out_trigger_unassign_all(CuTest *tc, struct wrtd_node *wrtd,
				      unsigned int chan)
{
	struct wrtd_trig_id id;
	int k, w;

	id.source_port = chan;
	for (k = MAX_SYSTEM; k > 0; k--) {
		for (w = 0; w < MAX_TRIGGER; w++) {
			id.system = k;
			id.trigger = w;
			test_out_trigger_unassign_one(tc, wrtd, chan, &id);
		}

	}
}

static void test_out_trigger_unassign(CuTest *tc)
{
	struct wrtd_output_trigger_state trig[15];
	struct wrtd_node *wrtd;
	unsigned int assigned;
	unsigned int i;
        int ret;

	wrtd = wrtd_open(1);

	for (i = 0; i < FD_NUM_CHANNELS; i++) {
                memset(&trig, 0, sizeof(trig));
                ret = wrtd_out_trig_get_all(wrtd, i, trig, MAX_ENTRY_CHAN);
		CuAssertIntEquals(tc, MAX_ENTRY_CHAN, ret);
		test_out_trigger_unassign_all(tc, wrtd, i);
		CuAssertIntEquals(tc, 0, wrtd_out_has_trigger(wrtd, i, NULL,
						       &assigned));
		CuAssertIntEquals(tc, 0, assigned);

		memset(&trig, 0, sizeof(trig));
                ret = wrtd_out_trig_get_all(wrtd, i, trig, MAX_ENTRY_CHAN);
		CuAssertIntEquals(tc, 0, ret);
	}

	wrtd_close(wrtd);
}

static void test_out_trigger_unassign_cond_all(CuTest *tc, struct wrtd_node *wrtd,
					   unsigned int chan)
{
	struct wrtd_trig_id id;
	int k, w;

	id.source_port = chan;
	for (k = 0; k < 2; k++) {
		for (w = 0; w < 2; w++) {
			id.system = k;
			id.trigger = w;
			test_out_trigger_unassign_one(tc, wrtd, chan, &id);
		}
	}
}

static void test_out_trigger_unassign_cond(CuTest *tc)
{
	struct wrtd_output_trigger_state trig[MAX_ENTRY_CHAN];
	struct wrtd_node *wrtd;
	unsigned int assigned;
	unsigned int i;
        int ret;

	wrtd = wrtd_open(1);

	for (i = 0; i < FD_NUM_CHANNELS; i++) {
		memset(&trig, 0, sizeof(ret));
                ret = wrtd_out_trig_get_all(wrtd, i, trig, MAX_ENTRY_CHAN);
		CuAssertIntEquals(tc, MAX_ENTRY_CHAN, ret);
		test_out_trigger_unassign_cond_all(tc, wrtd, i);
		CuAssertIntEquals(tc, 0, wrtd_out_has_trigger(wrtd, i, NULL,
						       &assigned));
		CuAssertIntEquals(tc, 0, assigned);

		memset(&trig, 0, sizeof(trig));
                ret = wrtd_out_trig_get_all(wrtd, i, trig, MAX_ENTRY_CHAN);
		CuAssertIntEquals(tc, 0, ret);
	}

	wrtd_close(wrtd);
}

static void test_out_pulse_width(CuTest *tc)
{
	struct wrtd_output_state st;
	struct wrtd_node *wrtd;
	unsigned int i, k;
	uint64_t ps;

	wrtd = wrtd_open(1);
	for (i = 0; i < FD_NUM_CHANNELS; i++) {
		CuAssertIntEquals(tc, -1, wrtd_out_pulse_width_set(wrtd, i, 0));
		CuAssertIntEquals(tc, errno, EWRTD_INVALID_PULSE_WIDTH);
		CuAssertIntEquals(tc, -1, wrtd_out_pulse_width_set(wrtd, i, 1000000000001ULL));
		CuAssertIntEquals(tc, errno, EWRTD_INVALID_PULSE_WIDTH);
		for (k = 250000; k < FD_NUM_CHANNELS; k+=10000) {
			CuAssertIntEquals(tc, 0, wrtd_out_pulse_width_set(wrtd, i, k));
			CuAssertIntEquals(tc, 0, wrtd_out_state_get(wrtd, i, &st));
			wrtd_ts_to_pico(&st.pulse_width, &ps);
			CuAssertIntEquals(tc, k, ps);
		}
	}
	wrtd_close(wrtd);
}

static void test_out_dead_time(CuTest *tc)
{
	struct wrtd_output_state st;
	struct wrtd_node *wrtd;
	unsigned int i, k;
	uint64_t ps;

	wrtd = wrtd_open(1);
	for (i = 0; i < FD_NUM_CHANNELS; i++) {
		for (k = 0; k < FD_NUM_CHANNELS; k+=10000) {
			CuAssertIntEquals(tc, 0, wrtd_out_dead_time_set(wrtd, i, k));
			CuAssertIntEquals(tc, 0, wrtd_out_state_get(wrtd, i, &st));
			wrtd_ts_to_pico(&st.dead_time, &ps);
			CuAssertIntEquals(tc, k, ps);
		}
	}
	wrtd_close(wrtd);
}

#define RANGE 10
static void test_out_trigger_delay(CuTest *tc)
{
	struct wrtd_node *wrtd;
	struct wrtd_output_trigger_state trig;
	struct wrtd_trigger_handle h[FD_NUM_CHANNELS];
	struct wrtd_trig_id id = {9,9,9};
	char msg[128];
	uint64_t ps, min, max;
	int i, k;
        int ret;

	wrtd = wrtd_open(1);
	for (i = 0; i < FD_NUM_CHANNELS; i++) {
		id.source_port = i;
                ret = wrtd_out_trig_assign(wrtd, i, &h[i], &id, NULL);
		CuAssertIntEquals(tc, 0, ret);
                ret = wrtd_out_trig_delay_set(wrtd, &h[i], 999999999999);
		CuAssertIntEquals(tc, -1, ret);
		CuAssertIntEquals(tc, EWRTD_INVALID_DELAY, errno);
		for (k = 1; k < 0x8000000; k <<= 1) {
                        ret = wrtd_out_trig_delay_set(wrtd, &h[i], k);
			CuAssertIntEquals(tc, 0, ret);
                        ret = wrtd_out_trig_state_get_by_handle(wrtd, &h[i], &trig);
			CuAssertIntEquals(tc, 0, ret);
			wrtd_ts_to_pico(&trig.delay_trig, &ps);
			/* Conversion is not precise, some approximation may
			   happen */
			min = k < RANGE ? 0 : k - RANGE;
			max = k + RANGE;
			sprintf(msg,
				"Assert failed - chan %d - %"PRIu64" not in range [%"PRIu64", %"PRIu64"]",
				i, ps, min, max);
			CuAssertTrue_Msg(tc, msg, ps >= min && ps <= max);
		}

	}

	for (i = 0; i < FD_NUM_CHANNELS; i++)
		wrtd_out_trig_unassign(wrtd, &h[i]);

	wrtd_close(wrtd);
}

static void test_out_trigger_enable(CuTest *tc)
{
	struct wrtd_node *wrtd;
	struct wrtd_output_trigger_state trig;
	struct wrtd_trigger_handle h;
	struct wrtd_trig_id id = {9,9,9};
	int i, k;

	wrtd = wrtd_open(1);
	for (i = 0; i < FD_NUM_CHANNELS; i++) {
		CuAssertIntEquals(tc, 0, wrtd_out_trig_assign(wrtd, i, &h, &id, NULL));
		for (k = 0; k < 10; k++) {
			/* Enable */
			CuAssertIntEquals(tc, 0, wrtd_out_trig_enable(wrtd, &h, 1));
			CuAssertIntEquals(tc, 0, wrtd_out_trig_state_get_by_handle(wrtd, &h,
									    &trig));
			CuAssertTrue(tc, trig.enabled);

			/* Disable */
			CuAssertIntEquals(tc, 0, wrtd_out_trig_enable(wrtd, &h, 0));
			CuAssertIntEquals(tc, 0, wrtd_out_trig_state_get_by_handle(wrtd, &h,
									    &trig));
			CuAssertIntEquals(tc, 0, trig.enabled);
		}
		wrtd_out_trig_unassign(wrtd, &h);
	}
	wrtd_close(wrtd);
}


static void test_out_trigger_mode(CuTest *tc)
{
	struct wrtd_node *wrtd;
	struct wrtd_output_state st;
	int i;

	wrtd = wrtd_open(1);
	for (i = 0; i < FD_NUM_CHANNELS; i++) {
		CuAssertIntEquals(tc, 0, wrtd_out_trigger_mode_set(wrtd, i,
							    WRTD_TRIGGER_MODE_SINGLE));
		CuAssertIntEquals(tc, 0, wrtd_out_state_get(wrtd, i, &st));
		CuAssertIntEquals(tc, WRTD_TRIGGER_MODE_SINGLE, st.mode);

		CuAssertIntEquals(tc, 0, wrtd_out_trigger_mode_set(wrtd, i,
							    WRTD_TRIGGER_MODE_AUTO));
		CuAssertIntEquals(tc, 0, wrtd_out_state_get(wrtd, i, &st));
		CuAssertIntEquals(tc, WRTD_TRIGGER_MODE_AUTO, st.mode);

		CuAssertIntEquals(tc, 0, wrtd_out_trigger_mode_set(wrtd, i,
							    WRTD_TRIGGER_MODE_SINGLE));
		CuAssertIntEquals(tc, 0, wrtd_out_state_get(wrtd, i, &st));
		CuAssertIntEquals(tc, WRTD_TRIGGER_MODE_SINGLE, st.mode);

	}
	wrtd_close(wrtd);
}

static void test_out_reset_counters(CuTest *tc)
{
	struct wrtd_node *wrtd;
	struct wrtd_output_state st;
	int i;

	wrtd = wrtd_open(1);
	for (i = 0; i < FD_NUM_CHANNELS; i++) {
		CuAssertIntEquals(tc, 0, wrtd_out_counters_reset(wrtd, i));
		CuAssertIntEquals(tc, 0, wrtd_out_state_get(wrtd, i, &st));
		CuAssertIntEquals(tc, 0, st.missed_pulses_late);
		CuAssertIntEquals(tc, 0, st.missed_pulses_deadtime);
		CuAssertIntEquals(tc, 0, st.missed_pulses_overflow);
		CuAssertIntEquals(tc, 0, st.missed_pulses_no_timing);
		CuAssertTrue(tc, !(st.flags & WRTD_LAST_VALID));
	}
	wrtd_close(wrtd);
}

CuSuite *wrtd_ut_out_suite_get(void)
{
	CuSuite *suite = CuSuiteNew();

	SUITE_ADD_TEST(suite, test_out_ping);
	SUITE_ADD_TEST(suite, test_out_channels_enable);
	SUITE_ADD_TEST(suite, test_out_channels_disable);
	SUITE_ADD_TEST(suite, test_out_channels_arm);
	SUITE_ADD_TEST(suite, test_out_channels_disarm);
	SUITE_ADD_TEST(suite, test_out_trigger_clear);
	SUITE_ADD_TEST(suite, test_out_trigger_assign);
	SUITE_ADD_TEST(suite, test_out_trigger_unassign);
	SUITE_ADD_TEST(suite, test_out_pulse_width);
	SUITE_ADD_TEST(suite, test_out_dead_time);
	SUITE_ADD_TEST(suite, test_out_trigger_delay);
	SUITE_ADD_TEST(suite, test_out_trigger_enable);
	SUITE_ADD_TEST(suite, test_out_trigger_mode);
	SUITE_ADD_TEST(suite, test_out_reset_counters);
	SUITE_ADD_TEST(suite, test_out_trigger_assign_cond);
	SUITE_ADD_TEST(suite, test_out_trigger_unassign_cond);

	return suite;
}
